# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .. import GenericFitter, GenericFitterTool, FunctionContainer
from ...data import DataSelection
from ...utils import DragPoint, DragPointManager, BlitManager, DragPointCollection


class DragLineManager(DragPointCollection):
    """Collection of DragPoints representing a straight line."""
    
    @staticmethod
    def function(x, m, n):
        """Straight line function.

        Parameters:
            x (float):
                independent variable.
            m (float):
                slope.
            n (float):
                value at `x=0`.

        Returns:
            (float):
                `f(x) = m*x+n`
        """
        return m*x + n

    @staticmethod
    def gradient(x, m, n):
        """Straigth line gradient.
        
        Parameters:
            x (float):
                independent variable.
            m (float):
                slope.
            n (float):
                value at 'x = 0'.
        
        Returns:
            (np.array):
                ' ( x, 1 )'
        """
        dfdm = x 
        dfdn = 1
        return np.array([[dfdm], [dfdn]])
    
    @staticmethod
    def get_args_length():
        """Gets number of arguments of `function`.

        Returns:
            (int): Number of arguments of `function`.
        """
        return 2
    
    def __init__(self, dragpoints: list[DragPoint], blit_manager: BlitManager):
        """Line between multiple DragPoints. Updates with them.

        Parameters:
            dragpoints (list[DragPoint]): line vertices.
            blit_manager (BlitManager): used for automtic ploting.
        """
        super().__init__(dragpoints, blit_manager)
        self.update()
    
    def update(self, *args, **kargs):
        """Updates line data with DragObjects positions"""
        x0, x1 = self.get_xdata()
        m, n = self.get_args()
        
        if (m,n) == (0,0):
            self.poly.set_xdata(self.get_xdata_display())
            self.poly.set_ydata(self.get_ydata_display())

        # create x and y data
        dx = abs(x0-x1)*0.5
        x = np.linspace(min(x0,x1)-dx, max(x0,x1)+dx, 250)
        y = self.function(x, m, n)
        
        # from data coordinates to display coordinates
        xy = np.array((x,y)).T.reshape(-1, 2)
        x_data, y_data = self.set_xy(xy).T
        
        # set new data
        self.poly.set_xdata(x_data)
        self.poly.set_ydata(y_data)
 
    def get_args(self):
        """Gives linear function parameters.

        Returns:
            (Tuple[float, float]):
                `m` and `n` of `f(x)=m*x + n`.
        """
        x0, x1 = self.get_xdata()
        y0, y1 = self.get_ydata()

        if (x1-x0)==0:
            return 0.,0.
        m:float = (y1 - y0)/(x1 - x0)
        n:float = m*(-x1)+y1
        return m, n

class LineFitter(GenericFitter):
    """Linear function fitter."""
    name = 'linear'
    
    def __init__(self, app, data: DataSelection):
        """Linear fitter following function `f(x)=m*x + n`.

        Parameters:
            app (Fitter): 
                Main application.
            data (DataSelection): 
                Data to fit.
        """
        super().__init__(app, data)
        
        ## Create DragPoints and DragLines needed
        
        self.drag_points = [DragPoint(*self.ax.transAxes.transform((0.2,0.3)), None), 
                            DragPoint(*self.ax.transAxes.transform((0.8,0.7)), None)]
        self.drag_points_managers = [DragPointManager(p, self.app.blit_manager) for p in self.drag_points]
        self.fitter_drag_collection = DragLineManager(self.drag_points, self.app.blit_manager)
        self.function = self.fitter_drag_collection.function
        self.gradient = self.fitter_drag_collection.gradient
        
        ## Connect Line to Points change events
        self.drag_points_cids = [] # Connections ids for change events
        for dp in self.drag_points_managers:
            self.drag_points_cids.append(
                dp.connect(self.fitter_drag_collection.update)
            )
        
        ## Add created DragPoints and DragLines to BlitManager's artists
        self.app.blit_manager.artists.append(self.fitter_drag_collection)
        for dpm in self.drag_points_managers:
            self.app.blit_manager.artists.append(dpm)
        
        self.fig.canvas.draw_idle()
    

class LineTool(GenericFitterTool):
    """Toggles Line Tool."""
    
    # default_keymap = ''
    description = 'Line me please'

    def enable(self, *args):
        """Triggered when LineTool is enabled.
        Uses BlitManager for faster rendering of DragObjects.
        """
        super().enable()
        self.fitter = LineFitter(self.app, self.data)

    def disable(self, *args):
        """Triggered when LineTool is disabled.
        Removes DragObjects and disables BlitManager.
        """
        super().disable()
        # If extra cleaning is needed
        


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .. import GenericFitter, GenericFitterTool
from ...data import DataSelection
from ...utils import DragPoint, DragPointManager, BlitManager, DragPointCollection


class DragExponentialManager(DragPointCollection):
    """Collection of DragPoints representing a straight line."""
    
    @staticmethod 
    def function(x,a,b):
        """Exponential function.

        Parameters:
            x (float):
                independent variable.
            a (float):
                scales exponential function.
            b (float):
                scales x.

        Returns:
            (float):
                `f(x) = a*exp(b*x)`
        """
        return a*np.exp(b*x)

    @staticmethod 
    def gradient(x,a,b):
        """Exponential gradient.

        Parameters:
            x (float):
                independent variable.
            a (float):
                scales exponential function.
            b (float):
                scales x.

        Returns:
            (np.array):
                `( exp(b*x), a*x*exp(b*x) )`
        """
        dfda = np.exp(b*x)
        dfdb = a * x * np.exp(b * x) 

        return np.array([[dfda], [dfdb]])
    
    @staticmethod
    def get_args_length():
        """Gets number of arguments of `function`.

        Returns:
            (int): Number of arguments of `function`.
        """
        return 2
    
    def __init__(self,dragpoints: list[DragPoint],blit_manager: BlitManager):
        """Exponential line between 2 DragPoints. Updates with them.

        Args:
            dragpoints (list[DragPoint]): line vertices.
            blit_manager (BlitManager): used for automatic ploting.
        """ 
        super().__init__(dragpoints,blit_manager)
        # name points based on use 
        self.point_1 = self.dragpoints[0]
        self.point_2 = self.dragpoints[1]
        self.update()

    def update(self,*args,**kargs):
        """Updates line data with DragObjects positions"""
        p1_x, p2_x = self.get_xdata()
        a,b = self.get_args()

        #create x and y data of an exponential that moves across two poitns
        dx = abs(p1_x-p2_x) * 1.5 
        x = np.linspace(min(p1_x,p2_x)-dx,max(p1_x,p2_x)+dx,250)
        y = self.function(x, a, b)

        # from data coordinates to display coordinates
        xy = np.array((x,y)).T.reshape(-1,2)
        x_data, y_data = self.set_xy(xy).T 

        #set new data
        self.poly.set_xdata(x_data)
        self.poly.set_ydata(y_data)

    def get_args(self):
        """Gives exponential function parameters.

        Returns:
            (Tuple[float,float]):
                `a`, and `b` of `f(x) = a*exp(b*x)`
        """ 
        p1_x, p1_y = self.get_xy(*self.point_1.patch.get_center())
        p2_x, p2_y = self.get_xy(*self.point_2.patch.get_center())

        b = 1/(p1_x - p2_x) * np.log(p1_y/p2_y)

        a = p1_y / np.exp(b*p1_x)
        return a, b 

class ExponentialFitter(GenericFitter):
    """Exponential function fitter."""
    name = 'exponential'

    def __init__(self,app,data: DataSelection):
        """ Exponential fitter following function `f(x) = a*exp(b*x)`

        Parameters:
            app (Fitter): 
                Main application.
            data (DataSelection): 
                Data to fit.
        """

        super().__init__(app,data)

        ## Create DragPoints and DragLines needed

        self.drag_points = [DragPoint(*self.ax.transAxes.transform((0.4,0.2)), None),
                            DragPoint(*self.ax.transAxes.transform((0.3,0.5)), None)]
        self.drag_points_managers = [DragPointManager(p,self.app.blit_manager) for p in self.drag_points]
        self.fitter_drag_collection = DragExponentialManager(self.drag_points, self.app.blit_manager)
        self.function = self.fitter_drag_collection.function
        self.gradient = self.fitter_drag_collection.gradient

        ## Connect Exponential to Points change events

        self.drag_points_cids = [] #Connection ids for change events
        for dp in self.drag_points_managers:
            self.drag_points_cids.append(
                dp.connect(self.fitter_drag_collection.update)
            )

        ## Add created DragPoints and DragLines to BlitManager's artists
        self.app.blit_manager.artists.append(self.fitter_drag_collection)
        for dpm in self.drag_points_managers:
            self.app.blit_manager.artists.append(dpm)
        self.fig.canvas.draw_idle()

class ExponentialTool(GenericFitterTool):
    """ Toggles Exponential Tool."""

    # default_keymap = '' 
    description = 'Exponentiate me please'

    def enable(self,*args):
        """Triggered when ExponentialTool is enabled.
        Uses BLitManager for faster rendering of DragObjects.
        """ 
        super().enable()
        self.fitter = ExponentialFitter(self.app, self.data)

    def disable(self,*args):
        """ Triggered when ExponentialTool is disabled
        Removes DragObjects and disables BLitManager.
        """ 
        super().disable()


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ... import Fitter

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backend_tools import ToolToggleBase
from matplotlib.lines import Line2D
from matplotlib.widgets import Button
from scipy import optimize
from matplotlib.patches import Polygon
from operator import add

from ...data import DataSelection, DataContainer
from ...utils import DragPointCollection, FitResultContainer

class GenericFitter:
    """GenericFitter is a base implementation of a fit function.
    All fit functions must inherit GenericFitter."""
    
    name = "generic"

    @staticmethod
    def function(x,*args) -> float:
        """Fit function: `f(x,*args)=...`.

        Parameters:
            x (float):
                Independent variable.
            *args (list[float,...]):
                0, 1 or multiple arguments.
        Returns:
            (float):
                `f(x, *args)`
        """
        ...
    
    @staticmethod
    def gradient(x,*args) -> float:
        """Fit gradient: 'f(x,*args)=...'.

        Paremeters:
            x (float):
                Independent variable.
            *args (list[float,...]):
                0, 1 or multiple arguments.
        Returns:
            (float):
                'f(x,*args)'
        """
        ...
        
    def get_args_length(self):
        """Gets number of arguments of `function`.

        Returns:
            (int): Number of arguments of `function`.
        """
        return self.fitter_drag_collection.get_args_length()
    
    def __init__(self, app: Fitter, data: DataSelection):
        """Generic fitter constructor.

        Parameters:
            app (Fitter):
                Main application.
            data (DataSelection):
                Data to fit.
        """
        self.app = app
        self.fig = app.figure
        self.ax = app.ax
        self.data = data
        
        self.fitter_drag_collection: DragPointCollection
        
        # TODO: this may change when dedicated ui is implemented
        self.button_axes = plt.axes([0.81, 0.000001, 0.1, 0.055])
        self.button = Button(self.button_axes, "Fit",color="red")
        self.button.on_clicked(self.on_fit)
        
    def get_args(self):
        """Return arguments needed for `self.function`.

        Returns:
            (Tuple[float]):
                0, 1 or multiple arguments.
        """
        return self.fitter_drag_collection.get_args()

    def prop_errors(xdata,self):
        """Returns the error of the fit, given a gradient.

        Returns:
            (float):
                sigma
        """
        return np.sqrt( float( self.gradient(xdata,*self.fit[0]).T @ self.fit[1] @ self.gradient(xdata,*self.fit[0]) ))
        
    
    def on_fit(self, event):
        """Event for fit button.

        Parameters:
            event (Matplotlib event): 
                Not used
        """

        # If there is not data selected use all data
        xdata, ydata = self.data.get_selected()
        xerr, yerr = self.data.get_selected_errors()
        if np.sum(self.data.indexes_used)==0:
            xdata, ydata = self.data.xdata.copy(), self.data.ydata.copy()
        
        self.fit = optimize.curve_fit(self.function, xdata, ydata, p0=self.get_args(), full_output=True, sigma=yerr)
        fit_result = FitResultContainer(self.data.copy(), self, self.fit)
        
        # Plot fit line in background, and the confidence interval
        with self.app.blit_manager.disabled():
            self.fit_line = Line2D(xdata, fit_result.evaluate(xdata), linestyle='--', color='purple')
            
            verts_positive, verts_negative = fit_result.error_verts()
            if verts_positive and verts_negative:
                self.fit_fill = Polygon(verts_positive + list(reversed(verts_negative)),facecolor='red',edgecolor='None',alpha=0.3)
                self.ax.add_artist(self.fit_fill)
                self.ax.draw_artist(self.fit_fill)

            self.ax.add_artist(self.fit_line)
            self.ax.draw_artist(self.fit_line)
            
       
       # Redraw plot to show line     
        self.app.blit_manager.draw()

        # Save fit in app
        self.app._add_fit(fit_result)
        
    def delete(self):
        """Remove trigger. Used when tool is disabled."""
        try:
            del self.button
            self.button_axes.remove()
        
            # Remove artists in order to clean canvas
            for pm in self.drag_points_managers:
                pm.dragpoint.remove()
                self.app.blit_manager.artists.remove(pm)

            self.fitter_drag_collection.remove()
            self.app.blit_manager.artists.remove(self.fitter_drag_collection) 
        
        except AttributeError:
            pass

        
        

class GenericFitterTool(ToolToggleBase):
    """Toggles Generic Fitter Tool."""
    
    default_toggled = False 
    radio_group = "fitter"

    def __init__(self, *args, app: Fitter, data: DataSelection, **kwargs):
        """Creates a GenericFitterTool.

        Parameters:
            app (Fitter):
                Main application.
            data (DataSelection):
                Data selected.
        """
        self.app = app
        self.data = data
        self.fitter: GenericFitter
        super().__init__(*args, **kwargs)

    def enable(self, *args):
        """Triggered when GenericTool is enabled.
        Uses BlitManager for faster rendering of DragObjects.
        """
        self.app.blit_manager.enable()

    def disable(self, *args):
        """Triggered when GenericTool is disabled.
        Removes DragObjects and disables BlitManager.
        """
        self.fitter.delete()
        self.app.blit_manager.disable()
        self.app.figure.canvas.draw_idle()


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class BlitManager:
    """**Do not use this class unless you know what blitting is and you are familiar with the rest of the code.**"""
    def __init__(self, app):
        """Class for managing blitting. DragObjects must be appended to `self.artists`.
        BlitManager must be manualy enabled and disabled.
        `with` statements can be used to enable or disable blitting temporaly.

        Parameters:
            app (Fitter):
                Aplication using BlitManager.
        """
        
        self.app = app
        self.ax = app.ax
        self.canvas = app.figure.canvas
        
        self.artists = []

        self._bind_motion_ : int = -1
        self._enabled_ = False
        self.background = None
        self.draw_event_connection_id = None
        
    def get_background(self):
        """"Gets current background and saves it, used in blitting process."""
        for a in self.artists:
            a.poly.set_visible(False)
        
        for a in self.ax.artists:
            self.ax.draw_artists(a)
            
        self.canvas.blit()

        return self.canvas.copy_from_bbox(self.ax.bbox)
    
    def update_background(self):
        """Updates saved background, used in blitting process."""
        self.background = self.get_background()

    def draw(self, artists_visible=True):
        """Draws the canvas using blitting."""
        if self.background is None:
            self.update_background()
        self.canvas.restore_region(self.background)
        
        for a in self.artists:
            try: # if custom object
                a.update()
                a.poly.set_visible(artists_visible)
                self.ax.draw_artist(a.poly)
            except AttributeError: # if matplotlib artists
                a.set_visible(artists_visible)
                self.ax.draw_artis(a)
            
        self.canvas.blit(self.ax.bbox)
        
    def on_draw(self, event):
        """Trigger for draw event."""
        self.draw()
              
    def enable(self):
        """Enables BlitManager."""
        if not self._enabled_:
            self.update_background()
            self._enabled_ = True
            # self.draw_event_connection_id = self.canvas.mpl_connect('draw_event', self.on_draw)
            
    def disable(self):
        """Disables BlitManager."""
        if self._enabled_:
            self._enabled_ = False
            # self.canvas.mpl_disconnect(self.draw_event_connection_id)
            self.draw_event_connection_id = None
      
    def enabled(self):
        """Enables Blit Manager and returns itself."""
        self.enable()
        return self     
     
    def disabled(self):
        """Disables BlitManager, redraws without DragObjects and returns itself."""
        self.disable()
        self.draw(artists_visible=False)
        return self
    
    @property
    def bind_motion(self):
        return self._bind_motion_
    
    @bind_motion.setter
    def bind_motion(self, new_id: int):
        # if (new_id is None) or (self._bind_motion_ is not None):
        if self._bind_motion_ == -1:
            self._bind_motion_ = new_id
        elif new_id is None:
            self._bind_motion_ = -1
            
        
    
    def __enter__(self, *_):
        pass
        
    def __exit__(self, *_):
        if not self._enabled_:
            self.enable()
        else:
            self.disable()


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from matplotlib.lines import Line2D

from . import DragPoint, BlitManager

class DragPointCollection:
    """A collection of DragPoints used to implement complex interactive functions.
    All collections of DragPoints must inherit from DragPointCollection and implement the following methods:
    
    * function: `f(x, *args)` that returns a float.
    * update: updates `DragPointCollection.poly` with `DragPointCollection.dragpoints` positions.
    * get_args: returns arguments needed for function (`*args`). Must be derived from `DragPointCollection.dragpoints` positions.
    """

    @staticmethod
    def function(*args, **kargs): ... 
    @staticmethod
    def get_args_length():...
    def update(self, *args, **kargs):...
    def get_args(self):...
    
    # Common methods
    def __init__(self, dragpoints: list[DragPoint], blit_manager: BlitManager, *, linestyle='-', color='red'):
        """Collection of DragPoints. Used to implement more complicated DragObjects.
        
        Args:
            dragpoints (list[DragPoint]): collection vertices.
            blit_manager (BlitManager): used for automtic ploting.
        """
        
        self.dragpoints = dragpoints
        self.blit_manager = blit_manager
        
        self.ax = blit_manager.ax
        self.canvas = blit_manager.canvas
        
        self.poly = Line2D(
            self.get_xdata_display(),
            self.get_ydata_display(),
            linestyle=linestyle,
            color=color,
            transform=None
        )
        
        self.patch = self.blit_manager.ax.add_patch(self.poly)  
        
    def get_xy(self, *args):
        """Applies and returns correct transformation from display to data coordinates.
        
        Parameters:
            *args (List[float,float] | List[List[float,float],...]):
                Coordinates from display.
        Returns:
            (Tuple[float,float] | Tuple[Tuple[float,float],...]):
                Coordinates from data.
        """

        args  = args if len(args)==2 else args[0]
        return self.ax.transData.inverted().transform(args)
    
    def set_xy(self, *args):
        """Applies and returns correct transformation from data coordinates to display.
        
        Parameters:
            *args (List[float] | List[List[float,float]]):
                Coordinates from data.
        Returns:
            (Tuple[float] | Tuple[Tuple[float,float]]):
                Coordinates from display.
        """
        args  = args if len(args)==2 else args[0]
        return self.ax.transData.transform(args)
        
    def get_xdata_display(self):
        """Gets xdata from DragPoints in display coordinates.
        
        Returns:
            (Tuple[float]):
                x in display coordinates.
        """
        
        return [p.get_center()[0] for p in self.dragpoints]
    
    def get_ydata_display(self):
        """Gets ydata from DragPoints in display coordinates.
        
        Returns:
            (Tuple[float]):
                y in display coordinates.
        """
        return [p.get_center()[1] for p in self.dragpoints]
    
    def get_xdata(self):
        """Gets xdata from DragPoints in data coordinates.
        
        Returns:
            (Tuple[float]):
                x in data coordinates.
        """
        return [self.get_xy(*p.get_center())[0] for p in self.dragpoints]
    
    def get_ydata(self):
        """Gets ydata from DragPoints in data coordinates.
        
        Returns:
            (Tuple[float]):
                y in data coordinates.
        """
        return [self.get_xy(*p.get_center())[1] for p in self.dragpoints]
    
    def remove(self):
        """Removes the patch from the axes."""
        self.patch.remove()
        


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...function_constructor import FunctionBuilder

import numpy as np
    
from .generic_fitter import GenericFitter


class FunctionContainer:
    operations = {
        "+" : lambda f,g: f + g,
        "-" : lambda f,g: f - g,
        "*" : lambda f,g: f * g,
        "/" : lambda f,g: f / g,
        "//" : lambda f,g: f // g,
        "%" : lambda f,g: f % g,
        "**" : lambda f,g: f ** g
    }

    chain_rule = {
        "+" : lambda f,g,df,dg: df + dg,
        "-" : lambda f,g,df,dg: df - dg,
        "*" : lambda f,g,df,dg: f*dg + g*df,
        "/" : lambda f,g,df,dg: (g*df - dg*f) / (g*g),
        "//" : lambda f,g,df,dg: None,
        "%" : lambda f,g,df,dg: None,
        "**" : lambda f,g,df,dg: f**(g-1)*(g*df + f*np.log(f)*dg)
    }
    
    def __init__(self, left_fitter: GenericFitter, function_builder: FunctionBuilder):
        self.function_builder = function_builder
        self.left_fitter = left_fitter
        self.left_fitter_args_legth: int = 0
        self.right_fitter : FunctionContainer | None = None
        self.right_fitter_args_legth: int = 0
        self.operation : str
        
    def copy(self):
        return FunctionContainer(self.left_fitter, self.function_builder)
        
    def _set_right_fitter_(self, right_fitter: GenericFitter|FunctionContainer):
        if self == right_fitter:
            self.right_fitter = self.copy()
        elif isinstance(right_fitter, GenericFitter):
            self.right_fitter = FunctionContainer(right_fitter, self.function_builder)
        elif isinstance(right_fitter, FunctionContainer):
            self.right_fitter = right_fitter
            self.right_fitter.set_function_builder(self.function_builder)
        
    def __add__(self, right_fitter: GenericFitter|FunctionContainer):
        self._set_right_fitter_(right_fitter)
        self.operation = '+'
        return self
    def __sub__(self, right_fitter: GenericFitter|FunctionContainer):
        self._set_right_fitter_(right_fitter)
        self.operation = '-'
        return self
    def __mul__(self, right_fitter: GenericFitter|FunctionContainer):
        self._set_right_fitter_(right_fitter)
        self.operation = '*'
        return self
    def __truediv__(self, right_fitter: GenericFitter|FunctionContainer):
        self._set_right_fitter_(right_fitter)
        self.operation = '/'
        return self
    def __floordiv__(self, right_fitter: GenericFitter|FunctionContainer):
        self._set_right_fitter_(right_fitter)
        self.operation = '//'
        return self
    def __mod__(self, right_fitter: GenericFitter|FunctionContainer):
        self._set_right_fitter_(right_fitter)
        self.operation = '%'
        return self
    def __pow__(self, right_fitter: GenericFitter|FunctionContainer):
        self._set_right_fitter_(right_fitter)
        self.operation = '**'
        return self
        
    def function(self, x, *args):
        if self.right_fitter is not None:
            return self.operations[self.operation](
                self.left_fitter.function(x, *args[:self.left_fitter_args_legth]),
                self.right_fitter.function(x, *args[self.left_fitter_args_legth:])
            )
        else:
            return self.left_fitter.function(x, *args[:self.left_fitter_args_legth])

    def gradient(self, x, *args):
        if self.right_fitter is not None:
            f = self.left_fitter.function(x, *args[:self.left_fitter_args_legth])
            g = self.right_fitter.function(x, *args[self.left_fitter_args_legth:])
            df = np.array(self.left_fitter.gradient(x, *args[:self.left_fitter_args_legth]))
            dg = np.array(self.right_fitter.gradient(x, *args[self.left_fitter_args_legth:]))
            return self.chain_rule[self.operation](
                f,
                g,
                np.append(df, np.zeros(dg.shape[0])),
                np.append(np.zeros(df.shape[0]), dg)
            )
        else:
            return self.left_fitter.gradient(x, *args[:self.left_fitter_args_legth])

    def get_args(self):
        if self.right_fitter is not None:
            return (*self.left_fitter.get_args(),*self.right_fitter.get_args())
        return self.left_fitter.get_args()
    
    def get_args_length(self):
        return self.left_fitter.get_args_length() + \
            (self.right_fitter.get_args_length() if self.right_fitter is not None else 0)

    def build(self):
        return self.function_builder.build()
    
    def set_function_builder(self, function_builder: FunctionBuilder):
        self.function_builder = function_builder
        if self.right_fitter is not None:  
            self.right_fitter.set_function_builder(function_builder)
    
    def _build_(self, plot_update_function: function):
        self.left_fitter = self.left_fitter(self.function_builder.app, self.function_builder.data)
        self.left_fitter_args_legth =  self.left_fitter.get_args_length()
        
        for dp in self.left_fitter.drag_points_managers:
            self.left_fitter.drag_points_cids.append(
                dp.connect(plot_update_function)
            )
            
        if self.right_fitter is not None:
            self.right_fitter._build_(plot_update_function)
            self.right_fitter_args_legth = self.right_fitter.get_args_length()
    


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .. import GenericFitter, GenericFitterTool
from ...data import DataSelection
from ...utils import DragPoint, DragPointManager, BlitManager, DragPointCollection


class DragQuadraticManager(DragPointCollection):
    """Collection of DragPoints representing a quadratic line."""
    
    @staticmethod
    def function(x, a, b, c):
        """Quadratic function.

        Parameters:
            x (float):
                independent variable.
            a (float):
                x^2 coefficient.
            b (float):
                x^1 coefficient.
            c (float):
                constant coefficient.

        Returns:
            (float):
                `f(x)=a*x^2 + b*x + c`
        """
        return a*x*x + b*x + c

    @staticmethod
    def gradient(x, a, b, c):
        """Quadratic function gradient.

        Parameters:
            x (float):
                independent variable.
            a (float):
                x^2 coefficient.
            b (float):
                x^1 coefficient.
            c (float):
                constant coefficient.

        Returns:
            (np.array):
                `( x^2, x, 1)`
        """
        dfda = x**2
        dfdb = x
        dfdc = 1
        return np.array([[dfda], [dfdb], [dfdc]])
    
    @staticmethod
    def get_args_length():
        """Gets number of arguments of `function`.

        Returns:
            (int): Number of arguments of `function`.
        """
        return 3
    
    def __init__(self, dragpoints: list[DragPoint], blit_manager: BlitManager):
        """Quadratic line between 2 DragPoints. Updates with them.

        Args:
            dragpoints (list[DragPoint]): line vertices.
            blit_manager (BlitManager): used for automtic ploting.
        """
        super().__init__(dragpoints, blit_manager)
        # name points based on use
        self.center_point = self.dragpoints[0]
        self.lateral_point = self.dragpoints[1]
        self.update()
    
    def update(self, *args, **kargs):
        """Updates line data with DragObjects positions"""
        cp_x, lp_x = self.get_xdata()
        a,b,c = self.get_args()

        # create x and y data of quadratic line centered in center_point
        dx = abs(lp_x-cp_x)*1.5
        x = np.linspace(cp_x-dx, cp_x+dx, 250)
        y = self.function(x, a, b, c)
        
        # from data coordinates to display coordinates
        xy = np.array((x,y)).T.reshape(-1, 2)
        x_data, y_data = self.set_xy(xy).T
        
        # set new data
        self.poly.set_xdata(x_data)
        self.poly.set_ydata(y_data)
        
    def get_args(self):
        """Gives quadratic function parameters.

        Returns:
            (Tuple[float, float, float]):
                `a`, `b` and `c` of `f(x)=a*x^2 + b*x + c`.
        """
        cp_x, cp_y = self.get_xy(*self.center_point.patch.get_center())
        lp_x, lp_y = self.get_xy(*self.lateral_point.patch.get_center())
        a = (lp_y - cp_y)/np.square(lp_x-cp_x)
        b = -2*a*cp_x
        c = cp_y + a*cp_x*cp_x
        return a,b,c

class QuadraticFitter(GenericFitter):
    """Quadratic function fitter."""
    name = 'quadratic'
    
    def __init__(self, app, data: DataSelection):
        """Quadratic fitter following function `f(x)=a*x^2 + b*x + c`

        Parameters:
            app (Fitter): 
                Main application.
            data (DataSelection): 
                Data to fit.
        """
        super().__init__(app, data)
        
        ## Create DragPoints and DragLines needed
        
        self.drag_points = [DragPoint(*self.ax.transAxes.transform((0.5,0.2)), None), 
                            DragPoint(*self.ax.transAxes.transform((0.7,0.5)), None)]
        self.drag_points_managers = [DragPointManager(p, self.app.blit_manager) for p in self.drag_points]
        self.fitter_drag_collection = DragQuadraticManager(self.drag_points, self.app.blit_manager)
        self.function = self.fitter_drag_collection.function
        self.gradient = self.fitter_drag_collection.gradient
        
        ## Connect Quadratic to Points change events
        self.drag_points_cids = [] # Connections ids for change events
        for dp in self.drag_points_managers:
            self.drag_points_cids.append(
                dp.connect(self.fitter_drag_collection.update)
            )
        
        ## Add created DragPoints and DragLines to BlitManager's artists
        self.app.blit_manager.artists.append(self.fitter_drag_collection)
        for dpm in self.drag_points_managers:
            self.app.blit_manager.artists.append(dpm)
        
        self.fig.canvas.draw_idle()
    

class QuadraticTool(GenericFitterTool):
    """Toggles Quadratic Tool."""
    
    # default_keymap = ''
    description = 'Quadratic me please'

    def enable(self, *args):
        """Triggered when QuadraticTool is enabled.
        Uses BlitManager for faster rendering of DragObjects.
        """
        super().enable()
        self.fitter = QuadraticFitter(self.app, self.data)

    def disable(self, *args):
        """Triggered when QuadraticTool is disabled.
        Removes DragObjects and disables BlitManager.
        """
        super().disable()
        # If extra cleaning is needed


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .builder import PlotBuilder
    
class SpineBuilder:
    def __init__(self, plot_builder: PlotBuilder):
        """Spine builder. Used to modify plot spines.

        Args:
            plot_builder (itfit.plot.PlotBuilder): PlotBuilder instance.
        """
        self._plot_builder_ = plot_builder
        self._dict_: dict[str, object] = {}
    
    def start_left_spine(self):
        """Starts a left spine builder.

        Returns:
            (itift.plot.spines.leftSpineBuilder): left spine builder.
        """
        return leftSpineBuilder(self._plot_builder_, self)
        
    def start_right_spine(self):
        """Starts a right spine builder.

        Returns:
            (itift.plot.spines.rightSpineBuilder): right spine builder.
        """
        return rightSpineBuilder(self._plot_builder_, self)
        
    def start_top_spine(self):
        """Starts a top spine builder.

        Returns:
            (itift.plot.spines.topSpineBuilder): top spine builder.
        """
        return topSpineBuilder(self._plot_builder_, self)
        
    def start_bottom_spine(self):
        """Starts a bottom spine builder.

        Returns:
            (itift.plot.spines.bottomSpineBuilder): bottom spine builder.
        """
        return bottomSpineBuilder(self._plot_builder_, self)
        
    def end_spines(self):
        """Ends spine builder.

        Returns:
            (itfit.plot.PlotBuilder): Returns the PlotBuilder.
        """
        return self._plot_builder_

class GenericSpineBuilder:
    def __init__(self, plot_builder: PlotBuilder, spine_builder: SpineBuilder):
        """Spine builder. Used to modify plot spines.

        Args:
            plot_builder (itfit.plot.spines.GenericSpineBuilder): GenericSpineBuilder instance.
        """
        self._plot_builder_ = plot_builder
        self._spine_builder_ = spine_builder
        self._dict_: dict[str, object] = {}
        
    def alpha(self, alpha: float):
        """Sets spines alpha.

        Args:
            alpha (float): alpha value

        Returns:
            (itfit.plot.spines.GenericSpineBuilder): Returns itself.
        """
        self._dict_.update({'alpha':alpha})
        return self

    def visible(self):
        """Sets spine visible.

        Returns:
            (itfit.plot.spines.GenericSpineBuilder): Returns itself.
        """
        self._dict_.update({'visible':True})
        return self
    
    def invisible(self):
        """Sets spine invisible.

        Returns:
            (itfit.plot.spines.GenericSpineBuilder): Returns itself.
        """
        self._dict_.update({'visible':False})
        return self

    def color(self, color):
        """Sets the color of the spine.

        Args:
            color (str|tuple[float]): Color for the spine.

        Returns:
            (itfit.plot.spines.GenericSpineBuilder): Returns itself.
        """
        self._dict_.update({'color': color})
        return self

    def linestyle(self, linestyle):
        """Changes spine linestyle.

        Args:
            linestyle (str): Spine linestyle.

        Returns:
            (itfit.plot.spines.GenericSpineBuilder): Returns itself.
        """
        self._dict_.update({'linestyle': linestyle})
        return self

    def linewidth(self, linewidth):
        """Changes spine linewidth.

        Args:
            linewidth (float): Spine linewidth.

        Returns:
            (itfit.plot.spines.GenericSpineBuilder): Returns itself.
        """
        self._dict_.update({'linewidth': linewidth})
        return self
    
    def _end_spine_builder(self, spine):
        self._plot_builder_.ax.spines[spine].set(**self._dict_)
        return self._spine_builder_


class leftSpineBuilder(GenericSpineBuilder):
    def __init__(self, plot_builder, spine_builder):
        super().__init__(plot_builder, spine_builder)
        self._spine = "left"
        
    def end_left_spine(self):
        """Ends left spine builder.

        Returns:
            (itfit.plot.PlotBuilder): Returns the PlotBuilder.
        """
        return self._end_spine_builder(self._spine)
        
        
class rightSpineBuilder(GenericSpineBuilder):
    def __init__(self, plot_builder, spine_builder):
        super().__init__(plot_builder, spine_builder)
        self._spine = "right"
        
    def end_right_spine(self):
        """Ends right spine builder.

        Returns:
            (itfit.plot.PlotBuilder): Returns the PlotBuilder.
        """
        return self._end_spine_builder(self._spine)
        
        
class topSpineBuilder(GenericSpineBuilder):
    def __init__(self, plot_builder, spine_builder):
        super().__init__(plot_builder, spine_builder)
        self._spine = "top"
        
    def end_top_spine(self):
        """Ends top spne builder.

        Returns:
            (itfit.plot.PlotBuilder): Returns the PlotBuilder.
        """
        return self._end_spine_builder(self._spine)
        

class bottomSpineBuilder(GenericSpineBuilder):
    def __init__(self, plot_builder, spine_builder):
        super().__init__(plot_builder, spine_builder)
        self._spine = "bottom"
        
    def end_bottom_spine(self):
        """Ends bottom spine builder.

        Returns:
            (itfit.plot.PlotBuilder): Returns the PlotBuilder.
        """
        return self._end_spine_builder(self._spine)
        


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .. import GenericFitter, GenericFitterTool
from ...data import DataSelection
from ...utils import DragPoint, DragPointManager, BlitManager, DragPointCollection


class DragSineManager(DragPointCollection):
    """Collection of DragPoints representing a sine function."""
    
    @staticmethod
    def function(x, a, b, c, d):
        """Sine function.

        Parameters:
            x (float):
                independent variable.
            a (float):
                Amplitude of the wave.
            b (float):
                frequency of the wave.
            c (float):
                centre of the sine function.
            d (float):
                constant value around which the wave oscillates.

        Returns:
            (float):
                `f(x) = a*sin(b*x+c)+d`
        """
        return a*np.sin(b*x+c) + d 

    @staticmethod
    def gradient(x, a, b, c, d):
        """Sine gradient function.

        Parameters:
            x (float):
                independent variable.
            a (float):
                Amplitude of the wave.
            b (float):
                frequency of the wave.
            c (float):
                centre of the sine function.
            d (float):
                constant value around which the wave oscillates.

        Returns:
            (np,array):
                ( sin(b*x + c), a*x*cos(b*x+c), a*cod(b*x+c), 1)
        """
        dfda = np.sin(b*x + c)
        dfdb = a * x* np.cos(b*x +c)
        dfdc = a * np.cos(b*x + c)
        dfdd = 1
        return np.array([[dfda], [dfdb], [dfdc], [dfdd]])
        
    def get_args_length():
        """Gets number of arguments of `function`.

        Returns:
            (int): Number of arguments of `function`.
        """
        return 4
    
    def __init__(self, dragpoints: list[DragPoint], blit_manager: BlitManager):
        """Sine function between multiple DragPoints. Updates with them.

        Parameters:
            dragpoints (list[DragPoint]): wave points.
            blit_manager (BlitManager): used for automtic ploting.
        """
        super().__init__(dragpoints, blit_manager)
        self.point_1 = self.dragpoints[0]
        self.point_2 = self.dragpoints[1]
        self.update()
    
    def update(self, *args, **kargs):
        """Updates function data with DragObjects positions"""
        p1_x, p2_x = self.get_xdata()
        a,b,c,d = self.get_args()

        #create x and y data of trigonometric function that moves across two points
        dx = abs(p1_x-p2_x) * 1.5 
        x = np.linspace(min(p1_x,p2_x)-dx,max(p1_x,p2_x)+dx,250)
        y = self.function(x, a, b, c,d)
        
        # from data coordinates to display coordinates
        xy = np.array((x,y)).T.reshape(-1, 2)
        x_data, y_data = self.set_xy(xy).T
        
        # set new data
        self.poly.set_xdata(x_data)
        self.poly.set_ydata(y_data)
 
    def get_args(self):
        """Gives sine function parameters.

        Returns:
            (Tuple[float, float]):
                `a`, `b`, `c` and `d` of `f(x)=a*sin(b*x +c)+d`.
        """
        x0, x1 = self.get_xdata()
        y0, y1 = self.get_ydata()

        a = abs(y0-y1)
        b = np.pi/2 / (x0-x1)
        c = -b *x1 
        d = y1

        return a, b, c, d

class SineFitter(GenericFitter):
    """Sine function fitter."""
    name = 'sine'

    def __init__(self, app, data: DataSelection):
        """Sine fitter following function 'f(x) = a*sin(b*x+b)'.
        
        Parameters:
            app (Fitter):
                Main application.
            data (DataSelection):
                Data to fit.
        """
        super().__init__(app,data)

        ##Create DragPoints and DragLines needed

        self.drag_points = [DragPoint(*self.ax.transAxes.transform((0.2,0.3)), None), 
                            DragPoint(*self.ax.transAxes.transform((0.8,0.7)), None)]
        self.drag_points_managers = [DragPointManager(p, self.app.blit_manager) for p in self.drag_points]
        self.fitter_drag_collection = DragSineManager(self.drag_points, self.app.blit_manager)
        self.function = self.fitter_drag_collection.function
        self.gradient = self.fitter_drag_collection.gradient

        ## Connect Line to Points change events
        self.drag_points_cids = [] # Connections ids for change events
        for dp in self.drag_points_managers:
            self.drag_points_cids.append(
                dp.connect(self.fitter_drag_collection.update)
            )

        ## Add created DragPoints and DragLines to BlitManager's artists
        self.app.blit_manager.artists.append(self.fitter_drag_collection)
        for dpm in self.drag_points_managers:
            self.app.blit_manager.artists.append(dpm)
        
        self.fig.canvas.draw_idle()

class SineTool(GenericFitterTool):
    """Toggles SineTool."""

    # default_keymap = ''
    description = 'Sine me please'

    def enable(self,*args):
        """Triggered when SineTool is enabled.
        Uses BlitManager for faster rendering of DragObjects.
        """

        super().enable()
        self.fitter = SineFitter(self.app,self.data)
    
    def disable(self, *args):
        """Triggered when SineTool is disabled.
        Removes DragObjects and disables BlitManager.
        """

        super().disable()
        # If extra cleaning is needed


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .. import GenericFitter, GenericFitterTool
from ...data import DataSelection
from ...utils import DragPoint, DragPointManager, BlitManager, DragPointCollection


class DragLorentzianManager(DragPointCollection):
    @staticmethod
    def function(x,A,x0,FWHM):
        """Lorentzian function.

        Args:
            x (float): independent variable.
            A (float): scalar.
            x0 (float): maximum center.
            FWHM (float): full width at half maximum.

        Returns:
            (float): `f(x) = A·FWHM / (π ·(x-x0)² + (FWHM/2)²)`
        """
        
        return A/np.pi*(FWHM/2)/((x-x0)**2+(FWHM/2)**2)

    @staticmethod
    def gradient(x,A,x0,FWHM):
        """# TODO: make docstrings

        Args:
            x (float): _description_
            A (float): _description_
            x0 (float): _description_
            FWHM (float): _description_

        Returns:
            (np.array): _description_
        """
        dfdA = 1/np.pi*(FWHM/2)/((x-x0)**2+(FWHM/2)**2)
        dfdx0 = A/np.pi*(FWHM/2) *2 *(x-x0) / ((x-x0)**2+(FWHM/2)**2)**2
        dfdF = A/np.pi/2 * ((x-x0)**2+(FWHM/2)**2) -  A/np.pi*(FWHM/2) * FWHM / ((x-x0)**2+(FWHM/2)**2)**2

        return A/np.pi*(FWHM/2)/((x-x0)**2+(FWHM/2)**2)
    
    @staticmethod
    def get_args_length():
        """Gets number of arguments of `function`.

        Returns:
            (int): Number of arguments of `function`.
        """
        return 3
    
    def __init__(self,dragpoints: list[DragPoint],blit_manager: BlitManager):
        """Lorentzian line between 2 DragPoints. Updates with them.

        Args:
            dragpoints (list[DragPoint]): line vertices.
            blit_manager (BlitManager): used for automatic ploting.
        """

        super().__init__(dragpoints,blit_manager)
        # name points based on use
        self.peak = self.dragpoints[0]
        self.side = self.dragpoints[1]
        self.update()
        
        

    def update(self,*args,**kargs):
        """Updates line data with DragObjects positions"""

        peak_x, side_x = self.get_xdata()
        A,x0,FWHM = self.get_args()

        #create x and y data of Lorentzian line of a Lorentzian that moves across two points
        dx = abs(peak_x-side_x) * 1.5
        x = np.linspace(min(peak_x,side_x)-dx,max(peak_x,side_x)+dx,250)
        y = self.function(x,A,x0,FWHM)

        #from data coordinates to display coordinates
        xy = np.array((x,y)).T.reshape(-1,2)
        x_data, y_data = self.set_xy(xy).T

        #set new data
        self.poly.set_xdata(x_data)
        self.poly.set_ydata(y_data)

    def get_args(self):
        """Gives Lorentzian function parameters.

        Returns:
            (tuple[float,float,float]): `A`, `x0`, and `FWHM` of `f(x) = A/pi*(FWHM/2)/((x-x0)^2+(FWHM/2)^2)`.
        """
        peak_x, peak_y = self.get_xy(*self.peak.patch.get_center())
        side_x, side_y = self.get_xy(*self.side.patch.get_center())

        if (peak_y < side_y and peak_y > 0) or \
           (peak_y >= side_y and peak_y < 0):
                peak_x, peak_y = self.get_xy(*self.side.patch.get_center())
                side_x, side_y = self.get_xy(*self.peak.patch.get_center())
        
        x0 = peak_x
        FWHM = 2*abs(peak_x-side_x)
        A = peak_y*FWHM/2*np.pi


        return A,x0,FWHM

class LorentzianFitter(GenericFitter):
    """Lorentzian function fitter."""
    name = 'lorentzian'

    def __init__(self,app,data: DataSelection):
        """ Lorentzian fitter following function `f(x) = A/pi*(FWHM/2)/((x-x0)^2+(FWHM/2)^2)`

        Parameters:
            app (Fitter):
                Main application.
            data (DataSelection):
                Data to fit.
        """

        super().__init__(app,data)

        ## Create DragPoints and DragLines needed

        self.drag_points = [DragPoint(*self.ax.transAxes.transform((0.5,0.7)), None),
                            DragPoint(*self.ax.transAxes.transform((0.7,0.3)), None)]
        self.drag_points_managers = [DragPointManager(p,self.app.blit_manager) for p in self.drag_points]
        self.fitter_drag_collection = DragLorentzianManager(self.drag_points, self.app.blit_manager)
        self.function = self.fitter_drag_collection.function
        self.gradient = self.fitter_drag_collection.gradient

        ##Connect Lorentzian to Points change events
        self.drag_points_cids = [] #Connections ids for change events
        for dp in self.drag_points_managers:
            self.drag_points_cids.append(
                dp.connect(self.fitter_drag_collection.update)
            )
        self.drag_points_managers[1].add_restriction(self.res_fwhm)
        
        ## Add created DragPoints and DragLines to BlitManager's artists
        self.app.blit_manager.artists.append(self.fitter_drag_collection)
        for dpm in self.drag_points_managers:
            self.app.blit_manager.artists.append(dpm)
        
        self.fig.canvas.draw_idle()

    def res_fwhm(self, x, y):
        return x, self.drag_points_managers[0].set_xy(0,
            self.drag_points_managers[0].get_xy(  *self.drag_points[0].patch.get_center() )[1]/2
            )[1]


        
    
class LorentzianTool(GenericFitterTool):
    """Toggles Lorentzian Tool."""

    # default_keymap = ''
    description = 'Lorentz me please'

    def enable(self,*args):
        """Triggered when LorentzianTool is enabled,
        Uses BlitManager for faster rendering of DragObjects.
        """
        super().enable()
        self.fitter = LorentzianFitter(self.app,self.data)

    def disable(self,*args):
        """Triggered when LorentzianTool is disabled.
        Removes DragObjects and disables BlitManager.
        """
        super().disable()


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TODO"""
try:
    __ITFIT_IMPORTED__
except NameError:
    __ITFIT_IMPORTED__= False
    
__version__ = "0.1.0"

if not __ITFIT_IMPORTED__:
    from . import data
    from . import fit_functions
    from . import data_selectors
    from .fitter_app import Fitter
    from .function_constructor import FunctionBuilder
    
    
__ITFIT_IMPORTED__ = True


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
from matplotlib.collections import RegularPolyCollection
from matplotlib.axes import Axes

class DataContainer:
    """Container for data.
    """
    def __init__(self, xdata: list, ydata: list, yerr: list|None=None, xerr: list|None=None):
        """Creates a DataContainer.

        Parameters:
            xdata (list[float]):
                x data.
            ydata (list[float]):
                y data.
            yerr (list | None, optional): 
                Error en y data. Defaults to None.
            xerr (list | None, optional): 
                Error in x data. Defaults to None.
        """
        self.xdata = np.array(xdata).copy()
        self.ydata = np.array(ydata).copy()
        self.xerr  = np.array(xerr).copy() if xerr is not None else None
        self.yerr  = np.array(yerr).copy() if yerr is not None else None
        
    def length(self):
        """Returns lenght of data.

        Returns:
            (int):
                lenght of data.
        """
        return self.xdata.size
    
    def get_data(self):
        """Returns data. As list of tuples: `lenght x 2`.

        Returns:
            (tuple[tuple[float, float]]):
                Data stored.
        """
        return np.array((self.xdata, self.ydata)).T
    
    def get_errors(self):
        """Returns data errors. A list of tuples: `lenght x2`.

        Returns:
            (tuple[tuple[float,float]]): 
                Errors in data stored.
        """
        return np.array((self.xerr, self.yerr), dtype=object).T

        
class DataSelection(DataContainer):
    def __init__(self, xdata, ydata, yerr: list|None=None, xerr: list|None=None):
        super().__init__(xdata, ydata, yerr=yerr, xerr=xerr)
        self.indexes_used = np.ones(len(self.xdata), dtype=bool)  

        self._was_plotted: bool = False
        self.collection: RegularPolyCollection = None
        self._axes: Axes = None

    def create_selected_poly(self, ax: Axes):
        """Creates a poly collection of selected data. Adds it to the given ax.
        """
        if self._was_plotted:
            return
        self._was_plotted = True
        self._axes = ax

        self.collection_facecolors = self.get_colors(
            (0, 1, 0, 1), (1, 0, 0, 1))
        self.collection = RegularPolyCollection(
            6, sizes=(40,),
            facecolors=self.collection_facecolors,
            offsets=self.get_data(),
            offset_transform=self._axes.transData)
        
        self.collection_ = self._axes.add_collection(self.collection)

    def _update_poly(self):
        """Updates poly collection colors.
        """
        if self._was_plotted:
            self.collection.set_facecolors(self.get_colors((0, 1, 0, 1), (1, 0, 0, 1)))

    def select_all(self):
        """Selects all data.
        """
        self.indexes_used[:] = True
        self._update_poly()
        
    def select_none(self):
        """Unselect all data.
        """
        self.indexes_used[:] = False
        self._update_poly()

    def add_selection(self, indexes: list):
        """Adds `indexes` to `indexes_used`.

        Parameters:
            indexes (list): 
                list of index.
        """
        self.indexes_used[np.array(indexes)] = True
        self._update_poly()
        
    def selection(self, indexes):
        """Erase previous selected indexes. Adds `indexes` to `indexes_used`.

        Parameters:
            indexes (list):
                list of index.
        """
        self.indexes_used[:] = False
        self.add_selection(indexes)
        self._update_poly()
        
    def bool_selection(self, indexes_used):
        """Erase previous selected indexes. Sets new `indexes_used`.

        Parameters:
            indexes_used (list):
                list of booleans. True if index used, False otherwise.
        """
        self.indexes_used[:] = indexes_used[:]
        self._update_poly()
        
    def get_selected(self):
        """Returns the selected data.

        Returns:
            (tuple[tuple[float], tuple[float]]):
                tuple containing x and y selected data in arrays.
        """
        return self.xdata[self.indexes_used], self.ydata[self.indexes_used]
    
    def get_selected_errors(self):
        """Returns the selected data errors.

        Returns:
            (tuple[tuple[float], tuple[float]]):
                tuple containing x and y selected data errors in arrays.
        """
        return self.xerr[self.indexes_used] if self.xerr is not None else None, self.yerr[self.indexes_used] if self.yerr is not None else None
    
    def get_not_selected(self):
        """Returns the not selected data.

        Returns:
            (tuple[tuple[float], tuple[float]]):
                tuple containing x and y not selected data in arrays.
        """
        return self.xdata[~self.indexes_used], self.ydata[~self.indexes_used]
    
    def get_not_selected_errors(self):
        """Returns the not selected data errors.

        Returns:
            (tuple[tuple[float], tuple[float]]):
                tuple containing x and y not selected data errors in arrays.
        """
        return self.xdata[~self.indexes_used] if self.xerr is not None else None, self.ydata[~self.indexes_used] if self.yerr is not None else None

    def get_colors(self, color_in, color_out):
        """Returns a list of colours depending if same index data is selected or not.

        Parameters:
            color_in (tuple[float,float,float,float]):
                Colour for selected data.
            color_out (tuple[float,float,float,float]):
                Colour for unselected data.

        Returns:
            (tuple[tuple[float,float,float,float]]):
                A list of colours.
        """
        colors = np.zeros((self.length(),4))
        colors[self.indexes_used,:] = color_in[:]
        colors[~self.indexes_used,:] = color_out[:]
        return colors
    
    def copy(self):
        """Creates a copy of the data selection object.

        Returns:
            (DataSelection): A copy of the data selection object.
        """
        instance = DataSelection(self.xdata.copy(), 
                                 self.ydata.copy(), 
                                 self.yerr.copy() if self.yerr is not None else None, 
                                 self.xerr.copy() if self.xerr is not None else None)
        instance.indexes_used = self.indexes_used.copy()
        return instance
        
if __name__=='__main__':
    d = DataSelection(xdata=[0,1,2], ydata=[3,4,5])
    
    assert (d.xdata == np.array([0,1,2])).all()         , "x input error"
    assert (d.ydata == np.array([3,4,5])).all()         , "y input error"
    assert (d.length() == 3)                            , "data length error"
    assert (d.get_data() == np.array([[0,3],
                                      [1,4],
                                      [2,5]])).all()    , "get_data error"
    
    d.selection([0,1])
    c = d.get_colors((0,0,0,1), (1,0,0,0))
    assert (c == [[0.,0.,0.,1.],
                  [0.,0.,0.,1.],
                  [1.,0.,0.,0.]]).all()                 , "get_colors error"
    print("All tests OK")


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .builder import PlotBuilder

class LabelBuilder:
    def __init__(self, plot_builder: PlotBuilder):
        """Starts a LabelBuilder. Used to modify axis labels.

        Args:
            plot_builder (PlotBuilder): PlotBuilder to return when exit.
        """
        self._plot_builder_ = plot_builder
        
        
    def start_x_label(self, label):
        """Starts a x label builder.

        Args:
            label (str): x axis label.

        Returns:
            (itfit.plot.labels.xLabelBuilder): x label builder.
        """
        return xLabelBuilder(self._plot_builder_, self,  label)
        
    def start_y_label(self, label):
        """Starts a y label builder.

        Args:
            label (str): y axis label.

        Returns:
            (itfit.plot.labels.yLabelBuilder): y label builder.
        """
        return yLabelBuilder(self._plot_builder_, self, label)
    
    def start_title(self, title):
        """Starts a title label builder.

        Args:
            title (str): title label.

        Returns:
            (itfit.plot.labels.titleLabelBuilder): title label builder.
        """
        return titleLabelBuilder(self._plot_builder_, self, title)
        
    def end_labels(self):
        """Exits labels builder.

        Returns:
            (itft.plot.PlotBuilder): Returns the PlotBuilder.
        """
        return self._plot_builder_

class GenericLabelBuilder:
    def __init__(self, plot_builder: PlotBuilder, label: str):
        """Abstract label builder. Use xLabelBuilder, yLabelBuilder or titleLabelBuilder.

        Args:
            plot_builder (itfit.plot.PlotBuilder): PlotBuilder instance.
            label (str): Label string.
        """
        self._plot_builder_ = plot_builder
        self._label_ = label
        self._dict_: dict[str, object] = {}

    def fontsize(self, size):
        """Sets label fontsize.

        Args:
            size (float): Size of the label.

        Returns:
            (itfit.plot.labels.GenericLabelBuilder): Returns itself.
        """
        self._dict_.update({'fontsize': size})
        return self

    def fontweight(self, weight):
        """TODO:

        Args:
            weight (_type_): _description_

        Returns:
            (itfit.plot.labels.GenericLabelBuilder): Returns itself.
        """
        self._dict_.update({'fontweight': weight})
        return self

    def color(self, color):
        """Sets the color of the label.

        Args:
            color (str|tuple[float]): Color for the label.

        Returns:
            (itfit.plot.labels.GenericLabelBuilder): Returns itself.
        """
        self._dict_.update({'color': color})
        return self

    def verticalalignment(self, loc):
        """TODO:

        Args:
            loc (_type_): _description_

        Returns:
            (itfit.plot.labels.GenericLabelBuilder): Returns itself.
        """
        self._dict_.update({'verticalalignment': loc})
        return self

    def horizontalalignment(self, loc):
        """TODO

        Args:
            loc (_type_): _description_

        Returns:
            (itfit.plot.labels.GenericLabelBuilder): Returns itself.
        """
        self._dict_.update({'horizontalalignment': loc})
        return self

class xLabelBuilder(GenericLabelBuilder):
    """Specific implementation of GenericLabelBuilder for x label.
    """
    def __init__(self, plot_builder: PlotBuilder, label_builder: LabelBuilder, label: str):
        super().__init__(plot_builder, label)
        self._label_builder_ = label_builder

    def end_xlabel(self):
        """Ends x label builder.

        Returns:
            (itfit.plot.labels.LabelBuilder): Returns the LabelBuilder.
        """
        self._plot_builder_.ax.set_xlabel(self._label_, self._dict_)
        return self._label_builder_

class yLabelBuilder(GenericLabelBuilder):
    """Specific implementation of GenericLabelBuilder for y label.
    """
    def __init__(self, plot_builder, label_builder: LabelBuilder, label: str):
        super().__init__(plot_builder, label)
        self._label_builder_ = label_builder

    def end_ylabel(self):
        """Ends y label builder.

        Returns:
            (itfit.plot.labels.LabelBuilder): Returns the LabelBuilder.
        """
        self._plot_builder_.ax.set_ylabel(self._label_, self._dict_)
        return self._label_builder_

class titleLabelBuilder(GenericLabelBuilder):
    """Specific implementation of GenericLabelBuilder for the title label.
    """
    def __init__(self, plot_builder, label_builder: LabelBuilder, label: str):
        super().__init__(plot_builder, label)
        self._label_builder_ = label_builder

    def end_title(self):
        """Ends title label builder.

        Returns:
            (itfit.plot.labels.LabelBuilder): Returns the LabelBuilder.
        """
        self._plot_builder_.ax.set_title(self._label_, self._dict_)
        return self._label_builder_


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
from matplotlib import path
from matplotlib.backend_tools import ToolToggleBase
from matplotlib.collections import RegularPolyCollection
from matplotlib.widgets import Lasso

from ..data import DataSelection


class LassoManager:
    """
    """
    def __init__(self, app, data: DataSelection):
        """Creates a lasso selector and applies the selection to the given data.

        Parameters:
            app (Fitter):
                Main application.
            data (DataSelection): 
                Data to apply selection.
        """
        self.app = app
        self.axes = app.ax
        self.canvas = self.axes.figure.canvas
        self.data = data

        self.Nxy = len(data.xdata)

        self.data.select_all()

        self.data.select_none()

        self.data_ploted = self.axes.plot(
            data.xdata, data.ydata, color=(0, 0, 0, 0))

        self.cid = self.canvas.mpl_connect('button_press_event', self.on_press)

        self.data.create_selected_poly(self.axes)

    def callback(self, verts):
        p = path.Path(verts)
        ind = p.contains_points(self.data.get_data())

        self.data.bool_selection(ind)
        self.delete()

    def delete(self):
        if hasattr(self, "lasso"):
            self.canvas.draw_idle()
            self.canvas.widgetlock.release(self.lasso)
            del self.lasso
        plt.disconnect(self.cid)
        self.cid = None

    def on_press(self, event):
        if self.canvas.widgetlock.locked():
            return
        if event.inaxes is None:
            return
        self.lasso = Lasso(event.inaxes,
                           (event.xdata, event.ydata),
                           self.callback)
        # acquire a lock on the widget drawing
        self.canvas.widgetlock(self.lasso)


class LassoTool(ToolToggleBase):
    """Toggles Lasso Tool."""
    # default_keymap = ''
    description = 'Lasso me please'

    def __init__(self, *args, app, data: DataSelection, **kwargs):
        """Creates a lasso tool.

        Parameters:
            app (Fitter): 
                Main application.
            data (DataSelection): 
                Data to apply selection.
        """
        self.app = app
        self.data = data
        super().__init__(*args, **kwargs)

    def enable(self, *args):
        """Enables the lasso tool. Interaction is locked until mouse button is released.
        """
        self.lasso_manager = LassoManager(self.app, self.data)

    def disable(self, *args):
        """Disables the lasso tool. After tool unselect or data selection applied.
        """
        self.lasso_manager.delete()


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from matplotlib.patches import Circle
from matplotlib.artist import Artist
import numpy as np


class DragPoint:
    """Data containter for draggable points. In the future it may support form, size and colour change.
    """
    def __init__(self, x, y, style, *args, **kwargs):
        """Creates a patch in given display coordinates.

        Parameters:
            x (float):
                y position in display units (between 0 and 1).
            y (float):
                y position in display units (between 0 and 1).
            style : Any
                No used.
        """
        self._style = style
        self.restriction_callback = lambda x,y: (x,y)
        color = kwargs.pop("color", "black")
        zorder = kwargs.pop("zorder", 10)
        self.patch = Circle(np.array([x,y]), 10, color=color, zorder=zorder, **kwargs)
        
    def get_center(self):
        """Returns the center position in display coordinates.

        Returns:
            (Tuple[float,float]):
                Center in display coordinates.
        """
        return self.restriction_callback(*self.patch.get_center())
        
    def remove(self):
        """Removes the point from the figure."""
        self.patch.remove()
        
        

class DragPointManager:
    """Manages `DragPoints`: event connection, automatic replotting on change/update, blitting and restrictions.
    """
    def __init__(self, dragpoint: DragPoint, blit_manager):
        """Manages a DragPoint's BlitManager connection, callbacks on matplotlib events and automatic drawing.

        Parameters:
            dragpoint (DragPoint): contains patch.
            blit_manager (BlitManager): used for automtic ploting.
        """
        self.dragpoint = dragpoint
        self.blit_manager = blit_manager
        
        self.ax = blit_manager.ax
        self.canvas = blit_manager.canvas
        
        self.dragpoint.patch.set_transform(None)
        self.blit_manager.ax.add_patch(self.dragpoint.patch)
        
        self.poly = self.dragpoint.patch
        self.connection_callbacks = {}
        self.restricction_callback = lambda x,y: (x,y)

        self.canvas.mpl_connect('button_press_event', self.on_button_press)
        self.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.canvas.mpl_connect('button_release_event', self.on_button_release)
        self.canvas.mpl_connect('motion_notify_event', self.on_mouse_move)
        
        self._ind_ = None # Used for enabling mouse motion.

    def connect(self, function):
        """Connects a callback for change envents. 
        
        Parameters:
            function (callable):
                Function to be executed when the DragPoint updates. Must have signature `def f(event)`.    
        Returns:
            (Int):
                Connection id. Can be used in `DragPointManager.disconnect`.
        """
        cin = np.random.randint(1000000)
        self.connection_callbacks.update({cin: function})
        return cin
    
    def disconnect(self, cid):
        """Disconnects the callback with given `cid`.
        
        Parameters:
            cid (Int):
                Connection id.
        """

        if cid in self.connection_callbacks.keys():
            self.connection_callbacks.pop(cid)
            
    def add_restriction(self, function):
        """Adds a restriction to point movement. 
        
        Parameters:
            function (callable):
                Must have signature `def f(new_x,new_y): -> Tuple[float, float]`
        """
        self.restricction_callback = function
        self.dragpoint.restriction_callback = self.restricction_callback
    
    def remove_restriction(self):
        """Removes the restriction to point movement"""
        self.restricction_callback = lambda x,y:(x,y)
        self.dragpoint.restriction_callback = self.restricction_callback

    def get_xy(self, x, y):
        """Applies correct transformation from display to data coordinates.
        
        Parameters:
            x (float):
                x in display coordinates.
            y (float):
                y in display coordinates.  
        Returns:
            (Tuple[float,float]):
                x and y in data coordinates.
        """
        return self.ax.transData.inverted().transform((x,y))
    
    def set_xy(self, x, y):
        """Applies correct transformation from data coordinates to display
        
        Parameters:
            x (float): 
                x in data coordinates.
            y (float):
                y in data coordinates.
        Returns:
            (Tuple[float,float]):
                x and y in display coordinates.
        """
        return self.ax.transData.transform((x,y))

    def on_button_press(self, event):
        """Callback for mouse button presses"""
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        if self.blit_manager.bind_motion != -1:
            return

        x, y = event.xdata, event.ydata
        x, y = self.set_xy(x, y)
        if np.hypot(*(self.poly.center - np.array([x,y]))) < 1.5*self.poly.get_radius():
            self._ind_ = hash(self)
        
            self.blit_manager.bind_motion = self._ind_
    
    def on_button_release(self, event):
        """Callback for mouse button releases"""
        if event.button != 1:
            return

        if self.blit_manager.bind_motion == self._ind_:
            self.blit_manager.bind_motion = None
        self._ind_ = 0

    def on_key_press(self, event):
        """Callback for key presses"""
        #! This may be useful later.
        if not event.inaxes:
            return

    def on_mouse_move(self, event):
        """Callback for mouse movements"""
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        if self._ind_ == 0:
            return
        if self.blit_manager.bind_motion != self._ind_:
            return
        
        x, y = event.xdata, event.ydata
        x, y = self.restricction_callback(x, y)
        x, y = self.set_xy(x, y)

        prop = {'center': np.array([x,y])}
        Artist.update(self.poly, prop)
        for k,v in self.connection_callbacks.items():
            v(x,y)
        
        self.blit_manager.draw()

    def update(self):
        x, y = self.dragpoint.get_center()
        Artist.update(self.poly, {'center': np.array([x,y])})


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from . import Fitter
    from .data import DataSelection
    
from matplotlib.lines import Line2D
import numpy as np

from .fit_functions import GenericFitter
from .fit_functions.common import (GenericFitter, GenericFitterTool, FunctionContainer)


class FunctionBuilder:
    def __init__(self, app: Fitter):
        self.app = app
        self.data = self.app.data
        self.function_container: FunctionContainer
        
    def define(self, function: FunctionContainer):
        self.function_container = function
        self.function_container.set_function_builder(self)
        
    def get_args(self):
        return self.function_container.get_args()
        
    def get_args_length(self):
        return self.function_container.get_args_length()

        
    def build(self):
        self.function_container._build_(self.update)
        
        self.function = self.function_container.function
        self.gradient = self.function_container.gradient
        self.args_length = self.function_container.get_args_length()
        
        if not hasattr(self, "fitter_instance"):
            self.fitter_instance = GenericFitter(app=self.app, data=self.data)
            self.fitter_instance.function_container = self
            self.fitter_instance.function = self.function_container.function
            self.fitter_instance.gradient = self.function_container.gradient
            self.fitter_instance.get_args_length = self.get_args_length
            self.fitter_instance.get_args = self.get_args
        
            self.poly = Line2D(
                self.data.get_data()[:,0],
                self.data.get_data()[:,1],
                linestyle='--',
                color='black',
                transform=None
            )
            self.patch = self.app.ax.add_patch(self.poly)
            
            self.app.blit_manager.artists.append(self)
        
        return self
        
    def update(self, *_):
        args = self.function_container.get_args()
        _x_data = self.data.get_selected()[0]
        _length = min(len(_x_data)*3, 250)
        x = np.linspace(min(_x_data), max(_x_data), _length)
        y = self.function(x, *args)
        
        xy = np.array((x,y)).T.reshape(-1,2)
        x_data, y_data = self.app.ax.transData.transform(xy).T

        self.poly.set_xdata(x_data)
        self.poly.set_ydata(y_data)
        
        
    def get_custom_tool(self):
        return self.CustomTool(self)
        
    class CustomTool(GenericFitterTool):
        """Toggles a custom tool Tool."""

        # default_keymap = ''
        description = 'Custom tool'
        
        def __init__(self, function_builder: FunctionBuilder):
            self.function_builder = function_builder
            
        def __call__(self, *args, app: Fitter, data: DataSelection, **kwargs):
            super().__init__(*args, app=app, data=data, **kwargs)
            return self

        def enable(self,*args):
            """Triggered when CustomTool is enabled,
            Uses BlitManager for faster rendering of DragObjects.
            """

            super().enable()
            self.fitter = self.function_builder.build().fitter_instance

        def disable(self,*args):
            """Triggered when CustomTool is disabled.
            Removes DragObjects and disables BlitManager.
            """

            super().disable()

        
        @classmethod 
        # This seems to be needed when using python 3.10
        def __subclasses__(cls):
            return []



# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from matplotlib.lines import Line2D

    from .. import Fitter

import matplotlib.pyplot as plt


class FitSelector:
    def __init__(self, app: Fitter):
        """Allows to select fits inside a figure, via click.

        Args:
            app (Fitter): Fitter aplication source of fits and data.
        """
        self.app = app
        
        self.fig, self.ax = plt.subplots()
        x_data, y_data = self.app.data.get_data().T
        line_data, = self.ax.plot(x_data, y_data, '.', c='black', label="data")
        
        self._key : int|list[int]
        self._mode_: str
        self.fit_lines: list[Line2D] = [line_data]
        self.legend_to_lines_map: dict[Line2D, Line2D] = {}
        self.lines_to_hash: dict[Line2D, int] = {}
        
        for key, fit in self.app.fits.items():
            line, = self.ax.plot(fit.get_fit_xdata(), fit.get_fit_ydata(), label=f"{fit.fit_manager.name}:{key}")
            self.fit_lines.append(line)
            self.lines_to_hash.update({line : key})
            

    def set_multiple_selection_mode(self):
        """Starts multiple selection mode picker and sets title.
        """
        self.ax.set_title('Click on legend line to toggle line on/off')
        self._mode_ = 'm'
        legend = self.ax.legend()
        
        for legend_line, original_line in zip(legend.get_lines(), self.fit_lines):
            legend_line.set_picker(True)
            self.legend_to_lines_map[legend_line] = original_line
            
        self.fig.canvas.mpl_connect('pick_event', self.on_pick_multiple)
        self.fig.show()
    
    def set_single_selection_mode(self):
        """Starts single selection mode picker and sets title.
        """
        self.ax.set_title('Click line to select it')
        self._mode_ = 's'
        for line in self.fit_lines:
            line.set_picker(True)
            
        self.fig.canvas.mpl_connect('pick_event', self.on_pick_single)
        self.fig.show()
        
        
    def on_pick_multiple(self, event):
        """Pick event for multiple selection mode.
        """
        legline = event.artist
        origline = self.legend_to_lines_map[legline]
        visible = not origline.get_visible()
        origline.set_visible(visible)
        
        legline.set_alpha(1.0 if visible else 0.2)
        self.fig.canvas.draw()

    def on_pick_single(self, event):
        """Pick event for single selection mode.
        """
        k = self.lines_to_hash.get(event.artist)
        if isinstance(k, int):
            self._key = k
        self.fig.canvas.stop_event_loop()
        return 
        
    def get_selected_from_figure(self):
        """Gets the lines selected on multiple selection mode based on visibility.
        """
        result: list[int] = []
        for line in self.fit_lines:
            if line.get_visible():
                if val := self.lines_to_hash.get(line):
                    result.append(val) 
        self.fig.canvas.stop_event_loop()
        self._key = result

    def get_selected(self):
        """Returns the selected fit/s.

        Returns:
            (int, list[int]): Keys of fits selected.
        """
        if self._mode_ == 'm':
            self.get_selected_from_figure()
        plt.close(self.fig)
        return self._key
    
    def connect_select_one(self):
        """Starts fit selection figure and waits for it to finish.

        Returns:
            (itFit.FitSelector): returns itself.
        """
        self._key = False
        self.set_single_selection_mode()

        if len(self.app.fits.keys()) == 1:
            self._key = self.app._last_fit
            return self
        
        self.fig.show()
        self.fig.canvas.start_event_loop()
        return self

            
    def connect_select_multiple(self):
        """Starts fit selection figure and waits for it to finish.

        Returns:
            (itFit.FitSelector): returns itself.
        """
        self._key = False
        self.set_multiple_selection_mode()
        self.fig.show()
        self.fig.canvas.start_event_loop()
        return self
        


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from . import data, utils
    from .function_constructor import FunctionBuilder
    from matplotlib.figure import Figure
    from matplotlib.axes import Axes
    

import matplotlib.pyplot as plt

from .data import DataSelection
from .data_selectors import LassoTool
from . import fit_functions
from .utils import BlitManager, FitSelector
from .utils.fit_container import FitResultContainer
from .plot.builder import PlotBuilder

plt.rcParams['toolbar'] = 'toolmanager'

class Fitter:
    data : data.DataSelection
    figure : Figure
    ax : Axes
    fits : dict[int, utils.FitResultContainer]
    selections : dict
    blit_manager : utils.BlitManager
    _last_fit : int
    
    def __init__(self, xdata, ydata, yerr=None, xerr=None, *args, **kargs):
        self.data = DataSelection(xdata, ydata, yerr=yerr, xerr=xerr)
        self.figure = plt.figure()
        self.ax = self.figure.gca()
        self.fits: dict[int, FitResultContainer] = {}
        self.selections = {}
        self.blit_manager = BlitManager(self)
        self._last_fit: int|None = None
        self._data_was_plotted = False
    
    def __call__(self):
        if not self._data_was_plotted:
            self.data_line = self.ax.plot(self.data.xdata, self.data.ydata, '.-')
            self._data_was_plotted = True
        
        self.figure.canvas.manager.toolmanager.add_tool('Lasso', LassoTool, app=self,data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Lasso', 'fitter')
        
        self.figure.canvas.manager.toolmanager.add_tool('Line', fit_functions.linear.LineTool, app=self, data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Line', 'fitter')
        
        self.figure.canvas.manager.toolmanager.add_tool('Quadratic', fit_functions.quadratic.QuadraticTool, app=self, data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Quadratic', 'fitter')

        self.figure.canvas.manager.toolmanager.add_tool('Exponential', fit_functions.exponential.ExponentialTool, app=self, data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Exponential', 'fitter')

        self.figure.canvas.manager.toolmanager.add_tool('Gaussian', fit_functions.gaussian.GaussianTool, app=self,data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Gaussian', 'fitter')

        self.figure.canvas.manager.toolmanager.add_tool('Sine',  fit_functions.sine.SineTool, app=self,data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Sine', 'fitter')
        
        self.figure.canvas.manager.toolmanager.add_tool('Cosine',  fit_functions.cosine.CosineTool, app=self,data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Cosine', 'fitter')
        
        self.figure.canvas.manager.toolmanager.add_tool('Lorentzian', fit_functions.lorentzian.LorentzianTool, app=self,data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Lorentzian', 'fitter')
        
    def add_custom_fit_function(self, function_builder: FunctionBuilder):
        if not self._data_was_plotted:
            self.data_line = self.ax.plot(self.data.xdata, self.data.ydata)
            self._data_was_plotted = True
        
        self.figure.canvas.manager.toolmanager.add_tool('Custom tool', function_builder.get_custom_tool(), app=self,data=self.data)
        self.figure.canvas.manager.toolbar.add_tool('Custom tool', 'fitter')

    def _add_fit(self, fit: FitResultContainer):
        """Adds the fit to the application

        Parameters:
            fit (FitResultContainer):
                Fit to add
        """
        self._last_fit = hash(fit)
        self.fits.update({self._last_fit: fit})

    def get_single_fit_selector(self):
        """Stars a fit selector figure where you can select one fit.

        Returns:
            (itfit.utils.FitResultContainer): Fit result container of selected fit.
        """
        selector = FitSelector(self)
        selection = selector.connect_select_one().get_selected()
        return self.fits.get(selection)

    def get_fit_selector(self):
        """Stars a fit selector figure where you can select one or more fits.

        Returns:
            (itfit.utils.FitResultContainer, list[itfit.utils.FitResultContainer]): Fit result container of selected fits, list if multiple.
        """
        selector = FitSelector(self)
        selection = selector.connect_select_multiple().get_selected()
        if isinstance(selection, list):
            return [self.fits.get(k) for k in selection]
        return self.fits.get(selection)
        
    def get_last_fit(self):
        """Returns the last fit

        Returns:
            (FitResultContainer): Fit result container.
        """
        return self.fits.get(self._last_fit) if (self._last_fit is not None) else None
    
    def add_filter(self, filter: function):
        """Adds a filter for data selection. The signature must be as:
```py
lambda x,y : bool
```
or
```py
def foo(x,y) -> bool:
    return bool
```
        Parameters:
            filter (function): Filter function.
        """
        selection = filter(*(self.data.get_data().T))
        self.data.selection(selection)
        self.data.create_selected_poly(self.ax)
    
    
    def get_plot_builder(self):
        """Returns a itfit.plot.PlotBuilder instance. Used to ease plot creation.
        """
        
        return PlotBuilder(self)
    
    def default_plot_last_fit(self, xlabel: str="", ylabel: str="", title: str=""):
        """Plots last fit with default configuration:
```py
.plot_data(label="Data")\\
.with_errors()\\
.with_fit(label=fit.fit_manager.name.capitalize())\\
.xlabel(xlabel).ylabel(ylabel).title(title)\\       
.spines()\\
    .start_top_spine().invisible().end_top_spine()\\
    .start_right_spine().invisible().end_right_spine()\\
.end_spines()\\     
.grid().legend().tight_layout()
```

Parameters:
    xlabel (str): x label. Defaults to "".
    ylabel (str): y label. Defaults to "".
    title (str): title. Defaults to "".
      
Returns:
    (itfit.plot.PlotBuilder): PlotBuilder to continue plot customization.
        """
        fit = self.get_last_fit()
        if fit is None:
            raise Exception("At least one fit must me made before trying to plot a fit.")
        return PlotBuilder(self, fit)\
            .plot_data(label="Data")\
            .with_errors()\
            .with_fit(label=fit.fit_manager.name.capitalize())\
            .xlabel(xlabel).ylabel(ylabel).title(title)\
            .spines()\
                .start_top_spine().invisible().end_top_spine()\
                .start_right_spine().invisible().end_right_spine()\
            .end_spines()\
            .grid().legend().tight_layout()



# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""


"""
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..fitter_app import Fitter
    from ..utils import FitResultContainer 
    
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib import rcParams
from matplotlib import style as mpl_style
from matplotlib import pyplot as plt
from matplotlib.patches import Polygon

from .labels import LabelBuilder
from .spines import SpineBuilder
from ..utils.fit_selector import FitSelector

class PlotBuilder:
    """PlotBuilder is a class whose intent is to uase the user in plot customization. 
    Using PlotBuilder with jupyther/notebook is strongly recomended as changes are interactive and figures are preserved.
    ![image](example1.PNG)
    """


    def __init__(self, app: Fitter, **kargs):
        """_summary_

        Parameters:
            app (itfit.Fitter): Main application.
        """
        self.fig: Figure
        self.ax: Axes
        self.fit: FitResultContainer
        self.app: Fitter = app
        self._only_selected_cache_: bool

    def plot_fit(self, fmt='--', color='black', label='', only_selected_data:bool=False, **kargs):
        """Plots the fit line into the figure.

        Parameters:
            fmt (str, optional): Fit line format. Defaults to '--'.
            color (str, optional): Color for the line. Defaults to 'black'.
            label (str, optional): Label assigned to the artists. Defaults to ''.
            only_selected_data (bool, optional): If `True` the fit is only shown for data used in optimization. Defaults to `False`.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        self._start_()

        if len(self.app.fits.keys()) > 1:
            key: int = FitSelector(self.app).connect_select_one().get_selected()
            self.fit = self.app.fits.get(key)
        else:
            self.fit = self.app.get_last_fit()
        
        if only_selected_data:
            DATA = self.fit.get_fit_data_selected().T
        else:
            DATA = self.fit.get_fit_data().T
            
        self.ax.plot(*DATA, fmt, color=color, label=label, **kargs)
        return self

    def with_fit(self, fmt='--', color='black', label='', only_selected:bool = False, **kargs):
        """Same as itfit.plot.builder.PlotBuilder.plot_fit. Plots the fit line into the figure.

        Parameters:
            fmt (str, optional): Fit line format. Defaults to '--'.
            color (str, optional): Color for the line. Defaults to 'black'.
            label (str, optional): Label assigned to the artists. Defaults to ''.
            only_selected (bool, optional): Shows fit only on selected data range. Defaults to `False`.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        return self.plot_fit(fmt=fmt, color=color, label=label, **kargs)
    
    def plot_fit_errors(self, color: str|tuple[float]|None = None, edgecolor = 'None', alpha: float = 0.3, only_selected:str|bool = 'auto', **kargs):
        """Plots error shadow on fit. Must be called after fit has been plotted.

        Parameters:
            color (str | tuple[float] | None, optional): Fill color. Defaults to None.
            edgecolor (str, optional): Color of the error shadow edges. Defaults to 'None'.
            alpha (float, optional): Alpha value. Defaults to 0.3.
            only_selected (str | bool, optional): If False error will be shown for all data range. Defaults to 'auto'.

        Raises:
            Exception: If used prior to fit plotting.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        try:
         _only_selected_ = only_selected if only_selected == 'auto' else self._only_selected_cache_
        except AttributeError:
            raise Exception("Fit must be plotted prior to fit's error shadow.")
        
        verts_positive, verts_negative = self.fit.error_verts(only_selected=_only_selected_)
        
        self.fit_fill = Polygon(verts_positive + list(reversed(verts_negative)),facecolor=color, edgecolor=edgecolor, alpha=alpha, **kargs)
        self.ax.add_artist(self.fit_fill)
        self.ax.draw_artist(self.fit_fill)
        return self

    def plot_data(self, fmt='.', color=None, label='', only_selected:bool = False, yerr:bool = True, xerr:bool = True, **kargs):
        """Plots the data fitted into the figure.

        Parameters:
            fmt (str, optional): Data line format. Defaults to '.'.
            color (str, optional): Color for the line. Defaults to None.
            label (str, optional): Label assigned to the artists. Defaults to ''.
            only_selected (bool, optional): Shows only selected data. Defaults to `False`.
            yerr (bool, optional): Shows data errors in y coordinate if given. Defaults to `True`.
            xerr (bool, optional): Shows data errors in y coordinate if given. Defaults to `True`.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        self._start_()
        self._only_selected_cache_ = only_selected
        try:
            if only_selected:
                DATA = self.fit.get_data_selected().T
            else:
                DATA = self.fit.get_data().T
        except AttributeError:
            DATA = self.app.data.get_data().T
            
        self.ax.plot(*DATA, fmt, color=color, label=label, **kargs)
        return self

    def with_data(self, fmt='.', color=None, label='', only_selected:bool = False, yerr:bool = True, xerr:bool = True, **kargs):
        """Same as itfit.plot.builder.PlotBuilder.plot_data. Plots the data fitted into the figure.

        Parameters:
            fmt (str, optional): Data line format. Defaults to '.'.
            color (str, optional): Color for the line. Defaults to None.
            label (str, optional): Label assigned to the artists. Defaults to ''.
            only_selected (bool, optional): Shows only selected data. Defaults to `False`.
            yerr (bool, optional): Shows data errors in y coordinate if given. Defaults to `True`.
            xerr (bool, optional): Shows data errors in y coordinate if given. Defaults to `True`.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        return self.plot_data(fmt=fmt, color=color, label=label, only_selected=only_selected, yerr=yerr, xerr=xerr,**kargs)
    
    def plot_data_errors(self, axis:str|bool = 'y', ecolor: str|tuple[float]|None = None, elinewidth: float|None = None, capsize: float|None = 1, errorsevery: int|tuple[int, int] = 1, **kargs):
        """Plots error bars on data. Must be called after data has been plotted.

        Parameters:
            axis (str | bool, optional): Axis in which to plot error bars. Defaults to y.
            ecolor (str | tuple[float] | None, optional): Sets the color of the error bar. Defaults to None.
            elinewidth (float | None, optional): Sets the line width of the error bar. Defaults to None.
            capsize (float | None, optional): Controls the length of the error bar cap in points. Defaults to 1.
            errorsevery (int | tuple[int, int], optional): Draws error bars on a subset of the data. e.g. every=(3,4) then error bars would be on
        x[3], x[7], x[11].... Defaults to 1.

        Raises:
            Exception: If used prior to data plotting.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        try:
            if self._only_selected_cache_:
                DATA = self.fit.get_data_selected().T
                YDATA_ERROR = self.fit.get_ydata_errors_selected()
                XDATA_ERROR = self.fit.get_xdata_errors_selected()
            else:
                DATA = self.fit.get_data().T
                YDATA_ERROR = self.fit.get_ydata_errors()
                XDATA_ERROR = self.fit.get_xdata_errors()
        except AttributeError:
            raise Exception("Data must be plotted prior to data's error bars.")

        if (axis=='x' or axis==True) and XDATA_ERROR is not None:
            self.ax.errorbar(*DATA, XDATA_ERROR, fmt='None', capsize=capsize, ecolor=ecolor, elinewidth=elinewidth, errorevery=errorsevery, **kargs)
        if (axis=='y' or axis==True) and YDATA_ERROR is not None:
            self.ax.errorbar(*DATA, YDATA_ERROR, fmt='None', capsize=capsize, ecolor=ecolor, elinewidth=elinewidth, errorevery=errorsevery, **kargs)

        return self
    
    def with_errors(self, axis:str|bool = 'y', ecolor: str|tuple[float]|None = None, elinewidth: float|None = None, capsize: float|None = 1, errorsevery: int|tuple[int, int] = 1, **kargs):
        """Same as itfit.plot.builder.PlotBuilder.plot_data_errors. Plots error bars on data. Must be called after data has been plotted.

        Parameters:
            axis (str | bool, optional): Axis in which to plot error bars. Defaults to y.
            ecolor (str | tuple[float] | None, optional): Sets the color of the error bar. Defaults to None.
            elinewidth (float | None, optional): Sets the line width of the error bar. Defaults to None.
            capsize (float | None, optional): Controls the length of the error bar cap in points. Defaults to 1.
            errorsevery (int | tuple[int, int], optional): Draws error bars on a subset of the data. e.g. every=(3,4) then error bars would be on
        x[3], x[7], x[11].... Defaults to 1.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        return self.plot_data_errors(axis, ecolor, elinewidth, capsize, errorsevery, **kargs)
    
    def labels(self):
        """Starts labels builder. After calling it xlabel and ylabel can be accessed.

        Returns:
            (itfit.plot.labels.LabelBuilder): label builder.
        """
        return LabelBuilder(self)
    
    def title(self, title: str):
        """Shortcut to `.labels().start_title(title).end_title().end_labels()`.

        Parameters:
            title (str): Title string.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        return self.labels().start_title(title).end_title().end_labels()

    def xlabel(self, xlabel):
        """Shortcut to `.labels().start_x_label(xlabel).end_xlabel().end_labels()`.

        Parameters:
            xlabel (str): x label string.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        return self.labels().start_x_label(xlabel).end_xlabel().end_labels()

    def ylabel(self, ylabel):
        """Shortcut to `.labels().start_y_label(ylabel).end_ylabel().end_labels()`.

        Parameters:
            ylabel (str): y label string.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        return self.labels().start_y_label(ylabel).end_ylabel().end_labels()


    def legend(self, *args, **kargs):
        """Toggles the legend in the plot.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        self.ax.legend(*args, **kargs)
        return self

    def grid(self, *args, **kwargs):
        """Toggles the grid in the plot.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        self.ax.grid(*args, **kwargs)
        return self

    def set_xlim(self, left: float, right: float, **kargs):
        """Sets the left and right plot limits on x axis.
        
        Parameters:
            left (float): left limit.
            right (float): right limit.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        self.ax.set_xlim(left=left, right=right, **kargs)
        return self

    def set_ylim(self, bottom: float, top: float, **kargs):
        """Sets the bottom and top plot limits on y axis.
        
        Parameters:
            bottom (float): bottom limit.
            top (float): top limit.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        self.ax.set_ylim(bottom=bottom, top=top, **kargs)
        return self

    def tight_layout(self, **kargs):
        """Adjust the padding between and around subplots.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        self.fig.tight_layout(**kargs)
        return self

    def set_size(self, size, unit='cm'):
        """Sets the figure size, be default in centimeters.
        
        Parameters:
            size (float): left limit.
            unit (float): right limit. Default='cm'.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        factor = 1/2.54
        if unit == 'inch':
            factor = 1
        if unit == 'px':
            factor = 1/rcParams['figure.dpi']
        self.fig.set_size_inches(size[0]*factor, size[1]*factor, forward=True)
        return self

    def spines(self):
        """Starts spines builder.

        Returns:
            (itfit.plot.spines.SpineBuilder): spine builder.
        """
        return SpineBuilder(self)
    
    def style(self, style):
        """Sets the style. Must be executed befor `start`.

        Parameters:
            style (str): Style to use.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        mpl_style.use(style)
    
        self._start_(warn=True)
        return self
    
    def _start_(self, warn=False):
        """Creates the figure and the axes.

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to start building the plot.
        """
        if hasattr(self, "fig") and hasattr(self, "ax"):
            if warn:
                print("Warning: style changes will not affect an existent figure. Use `style` before ploting anything to the figure.")
            return self
        self.fig = plt.figure()
        self.ax = self.fig.gca()

    def save_fig(self, filename, transparent=False, **kargs):
        """Saves the figure with the given filename

        Returns:
            (itfit.plot.builder.PlotBuilder): Returns itself to continue building the plot.
        """
        self.fig.savefig(filename, transparent=transparent, **kargs)
        return self

    def show_inline(self):
        """Returns the figure. When used in Jupyter shows the figure in the cell output.

        Returns:
            (matplotlib.Figure): PlotBuilder's figure.
        """
        return self.fig


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""#Fitting functions
All fitting functions inherit from GenericFitter and follow the same structure.
To create a new function you have 2 options:

* Use FunctionBuilder to create function from already created ones.
* Write a new fitter using GenericFitter.

Here we will

"""
try:
    __FITTER_FIT_FUNCTIONS_IMPORTED__
except NameError:
    __FITTER_FIT_FUNCTIONS_IMPORTED__= False

if not __FITTER_FIT_FUNCTIONS_IMPORTED__:
    from .common.generic_fitter import GenericFitter, GenericFitterTool
    from .common.function_container import FunctionContainer
    
    from .linear import Line
    from .gaussian import Gaussian
    from .exponential import Exponential
    from .quadratic import Quadratic
    from .lorentzian import Lorentzian
    from .cosine import Cosine
    from .sine import Sine
    
__FITTER_FIT_FUNCTIONS_IMPORTED__ = True


# Copyright 2023 Unai Lería Fortea & Pablo Vizcaíno García

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..data import DataSelection
    from ..fit_functions.common import FunctionContainer, GenericFitter


import numpy as np


class FitResultContainer:
    def __init__(self, data: DataSelection, fit_manager: FunctionContainer|GenericFitter, scipy_result: dict):
        """_summary_

        Parameters:
            data (itfit.data.DataSelection):
                Data fitted.
            fit_manager (FunctionContainer|GenericFitter):
                Fit function used
            scipy_result (dict):
                Dictionary of `scipy.optimize.curve_fit` output.
        """
        self.data = data
        self.function = fit_manager.function
        self.gradient = fit_manager.gradient
        self.fit_manager = fit_manager
        self.scipy_output = {
            "popt" : scipy_result[0],
            "pcov" : scipy_result[1],
            "fvec" : scipy_result[2]["fvec"],
            "nfev" : scipy_result[2]["nfev"],
            "fjac" : scipy_result[2]["fjac"],
            "ipvt" : scipy_result[2]["ipvt"],
            "qtf"  : scipy_result[2]["qtf"],
            "mesg" : scipy_result[3],
            "ier"  : scipy_result[4] 
        }
        
    def get_parameters(self):
        """Gets the optimal fitting parameters found.

        Returns:
            (tuple[float]):
                tuple of parameters.
        """
        return self.scipy_output["popt"]
        
    def get_parameters_covariance(self):
        """Gets the parameters covariance matrix.

        Returns:
            (Ndarray(NxN)[float]):
                Parameters covariance matrix.
        """
        return self.scipy_output["pcov"]
        
    def get_parameters_errors(self):
        """Gets the square root of diagonal elements of the covariance matrix.

        Returns:
            (tuple[float]):
                Optimal fitting parameters standard error.
        """
        return np.sqrt(np.diag(self.get_parameters_covariance()))

    def get_xdata(self):
        """Gets the x component of all the data.

        Returns:
            (tuple[float]):
                X component of all the data.
        """
        return self.data.xdata
    
    def get_xdata_errors(self):
        """Gets the x component error in all the data.

        Returns:
            (tuple[float]):
                X component error of all the data.
        """
        return self.data.get_errors()[0]
        
    def get_ydata(self):
        """Gets the y component of all the data.

        Returns:
            (tuple[float]):
                Y component of all data.
        """
        return self.data.ydata
    
    def get_ydata_errors(self):
        """Gets the y component error in all the data.

        Returns:
            (tuple[float]):
                Y component error of all the data.
        """
        return self.data.get_errors()[1]
    
    def get_xdata_selected(self):
        """Gets the x component of the data used.

        Returns:
            (tuple[float]):
                X component of data used.
        """
        return self.data.get_selected()[0]
    
    def get_xdata_errors_selected(self):
        """Gets the x component error in data used.

        Returns:
            (tuple[float]):
                X component error of data used.
        """
        return self.data.get_selected_errors()[0]
    
    def get_ydata_selected(self):
        """Gets the y component of the data used.

        Returns:
            (tuple[float]):
                Y component of data used.
        """
        return self.data.get_selected()[1]
    
    def get_ydata_errors_selected(self):
        """Gets the y component error in data used.

        Returns:
            (tuple[float]):
                Y component error of data used.
        """
        return self.data.get_selected_errors()[1]
    
    def get_data(self):
        """Gets the all data.

        Returns:
            (tuple[tuple[float], tuple[float]]):
                All data.
        """
        return self.data.get_data()
    
    def get_data_selected(self):
        """Gets the data used.

        Returns:
            (tuple[tuple[float], tuple[float]]):
                Data used.
        """
        return np.array(self.data.get_selected()).T

    def prop_errors(self):
        """ Return the error of the fit, given a gradient of a function.

        Returns:
            (Tuple[float]):
                errors of the fit
        """
        try:
            x_array = self.get_fit_xdata()
            errors = np.zeros((len(x_array)))
            cov = self.get_parameters_covariance()

            for i,x in enumerate(x_array):
                grad = self.gradient(x, *self.get_parameters())
                errors[i] = np.sqrt( float(grad.T @ cov @ grad))
        except AttributeError:
            return None
        return errors


    def get_fit_xdata(self):
        """Gets the x component of the fit curve. Equal to get_xdata output.

        Returns:
            (tuple[float]):
                X component of fit curve. Equal to get_xdata output.
        """
        return self.get_xdata()
    
    def get_fit_ydata(self):
        """Gets the y coomponent of the fit curve.

        Returns:
            (tuple[float]):
                Y component of fit curve.
        """
        return self.function(self.get_xdata(), *self.get_parameters())
    
    def get_fit_xdata_selected(self):
        """Gets the x component of the fit curve for selected data interval. Equal to get_xdata_selected output.

        Returns:
            (tuple[float]):
                X component of fit curve. Equal to get_xdata output.
        """
        return self.get_xdata_selected()
    
    def get_fit_ydata_selected(self):
        """Gets the y coomponent of the fit curve for selected data interval.

        Returns:
            (tuple[float]):
                Y component of fit curve.
        """
        return self.function(self.get_xdata_selected(), *self.get_parameters())
    
    def get_fit_data(self):
        """Gets the fit curve data for all data.

        Returns:
            (tuple[tuple[float], tuple[float]]):
                Fit curve data.
        """
        return np.array((self.get_fit_xdata(), self.get_fit_ydata())).T
    
    def get_fit_data_selected(self):
        """Gets the fit curve data for selected data.

        Returns:
            (tuple[tuple[float], tuple[float]]):
                Fit curve data.
        """
        return np.array((self.get_fit_xdata_selected(), self.get_fit_ydata_selected())).T
    
    def get_message(self):
        """Gets scipy output `mesg` output.

        Returns:
            (str):
                Scipy output message.
        """
        return self.scipy_output["mesg"]
    
    def evaluate(self, x):
        """Evaluates the given `x` in the fitting function with the optimal parameters.

        Parameters:
            x (float): 
                Independent variable.

        Returns:
            y (float): 
                Dependent variable.
        """
        return self.function(x, *self.get_parameters())
    
    def error_verts(self, only_selected: bool=True):
        """Returns a tuple of two lists of points representing the error of the optimization. 
        Each lists size is Nx2 or None if errors in the function are not supported.

        Parameters:
            only_selected (bool):
                Only get errors points for selected data. Defaults to True.

        Returns:
            (tuple[list]|tuple[None]): Positive and negative error points.
        """
        xdata, _  = self.data.get_selected() if only_selected else self.data.get_data().T
        error_fit = self.prop_errors()
        if error_fit is not None:
            iy_pos = self.evaluate(xdata)+error_fit
            verts_positive = [(xdata[0],self.evaluate(xdata[0])), *zip(xdata,iy_pos), (xdata[len(xdata)-1],self.evaluate(xdata[len(xdata)-1]))]
            
            iy_neg = self.evaluate(xdata)-error_fit
            verts_negative = [(xdata[0],self.evaluate(xdata[0])), *zip(xdata,iy_neg), (xdata[len(xdata)-1],self.evaluate(xdata[len(xdata)-1]))]
            return verts_positive, verts_negative
        return (None, None)
                

    def __str__(self):
        TAB = "\t"
        NEX = "\n"

        return f"""ItFit FitResultContainer
Using fit function: {self.fit_manager.name}
Scipy result message: {self.get_message()}

Optimal parameters: 
{TAB}values: {self.get_parameters()}
{TAB}errors: {self.get_parameters_errors()}
{TAB}covariance:
{TAB}{TAB}[{(NEX + TAB*2 +" ").join([str(l) for l in self.get_parameters_covariance()])}]
"""
        
    def save(self, filename): # TODO:
        ...
    
    @classmethod    
    def load(cls, filename): # TODO:
        ...